package be.iminds.ilabt.jfed.highlevel.jobs;

import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.ExperimentPart;
import be.iminds.ilabt.jfed.experiment.SfaExperimentPart;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.jobs.parts.ExperimentPartController;
import be.iminds.ilabt.jfed.highlevel.jobs.parts.ExperimentPartControllerManager;
import be.iminds.ilabt.jfed.highlevel.jobs.parts.NonSfaExperimentPartController;
import be.iminds.ilabt.jfed.highlevel.jobs.sfa_parts.UpdateSfaExperimentPartUntilReadyStateSlice;
import be.iminds.ilabt.jfed.highlevel.jobs.states.JobStateFactory;
import be.iminds.ilabt.jfed.highlevel.jobs.states.ShareLanAtAuthorityPartState;
import be.iminds.ilabt.jfed.highlevel.model.InternalState;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.util.AggregateManagerWrapperFactory;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.RspecLink;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import java.util.*;
import java.util.stream.Collectors;

/**
 * This executes all actions that jFed needs to do AFTER an experiment is READY:
 *
 * - Sharing LANs (if specified in RSpec)
 * - Waiting for startup scripts (if wait method specified in RSpec)
 * - Distributing SSH keys (if specified in RSpec)
 * - Executing ansible (if ansible stuff specified in RSpec)
 * - Execute the dir, upload and execute steps of the ExperimentSpecification (if present)
 *
 * This does not:
 * - Test Connectivity  (should be done BEFORE this step)
 */
public class SetupSoftwareExperimentJob extends AbstractJob<Boolean> {
    private static final Logger LOG = LoggerFactory.getLogger(SetupSoftwareExperimentJob.class);

    private final JobStateFactory jobStateFactory;

    private final AggregateManagerWrapperFactory aggregateManagerWrapperFactory;
    private final ExperimentPartControllerManager epcManager;

    private Boolean success = null;

    public SetupSoftwareExperimentJob(Experiment experiment,
                                      HighLevelTaskFactory hltf,
                                      TaskThread tt,
                                      JobStateFactory jobStateFactory,
                                      AggregateManagerWrapperFactory aggregateManagerWrapperFactory,
                                      ExperimentPartControllerManager epcManager) {
        super("Setup Software for Experiment " + experiment.getName(), experiment, hltf, tt);
        this.jobStateFactory = jobStateFactory;
        this.aggregateManagerWrapperFactory = aggregateManagerWrapperFactory;
        this.epcManager = epcManager;

        assert experiment.getRequestedEndTime() != null;
    }

    @Override
    public Boolean execute() throws Exception {
        if (!hasSfaParts()) {
            if (experiment.getExperimentSpecification() != null) {
                experiment.getExperimentSpecificationLogger().firePostRSpec(experiment.getExperimentSpecification().getRspecs().get(0), false);
                experiment.getExperimentSpecificationLogger().firePostRspecAll(false, Collections.emptyList());
            }
            return true;
        }

        LOG.debug("SetupSoftwareExperimentJob started");

        //Assumption: everything is now successfully running and we have connectivity available

        if (experiment.getSliceOrNull() == null ||
                experiment.getSlice().getRequestRspec() == null ||
                experiment.getSlice().getRequestRspec().getModelRspec() == null) {
            LOG.warn("Problem with experiment.getSlice().getRequestRspec(). " +
                    "Will assume this is normal and ABORT software setup.");

            if (experiment.getExperimentSpecification() != null) {
                experiment.getExperimentSpecificationLogger().firePostRSpec(
                        experiment.getExperimentSpecification().getRspecs().get(0), true);
                experiment.getExperimentSpecificationLogger().firePostRspecAll(
                        true, getExperimentClientIds());
            }

            success = true;
            return true;
        }

        boolean hasAutoShareLans = false;
        Collection<? extends RspecLink> links = experiment.getSlice()
                .getRequestRspec().getModelRspec().getLinks();
        for (RspecLink link : links) {
            List<String> autoShareLanNames = link.getAutoShareLanNames();
            //noinspection ConstantConditions
            if (autoShareLanNames != null && !autoShareLanNames.isEmpty()) {
                hasAutoShareLans = true;
                break;
            }
        }
        if (hasAutoShareLans) {
            LOG.info("Experiment has auto shared LANs. Will share them.");
            final AutoShareLanPartsState autoShareLanPartsState = new AutoShareLanPartsState();
            setAndRunState(autoShareLanPartsState);

            if (autoShareLanPartsState.getStatus() != ExperimentTaskStatus.SUCCESS) {
                LOG.info("Failure sharing LANs");

                if (experiment.getExperimentSpecification() != null) {
                    experiment.getExperimentSpecificationLogger().firePostRSpec(
                            experiment.getExperimentSpecification().getRspecs().get(0), false);
                    experiment.getExperimentSpecificationLogger().firePostRspecAll(
                            false, getExperimentClientIds());
                }

                success = false;
                return false;
            }

            LOG.debug("Experiment auto shared LANs have been shared. Will wait for ready again.");
            final WaitForReadyExperimentJob.UpdateSfaPartsUntilReadyState updateUntilReadyState2 =
                    new WaitForReadyExperimentJob.UpdateSfaPartsUntilReadyState(
                            jobStateFactory, this);
            setAndRunState(updateUntilReadyState2);

            if (updateUntilReadyState2.isPermanentFailure()) {
                LOG.info("WaitForPartsReadyState (after auto shared LANs) reported permanent failure");

                if (experiment.getExperimentSpecification() != null) {
                    experiment.getExperimentSpecificationLogger().firePostRSpec(
                            experiment.getExperimentSpecification().getRspecs().get(0), false);
                    experiment.getExperimentSpecificationLogger().firePostRspecAll(
                            false, getExperimentClientIds());
                }

                success = false;
                return false;
            }

            LOG.info("All is ready after sharing LANs -> all ready");
        } else
            LOG.info("Experiment has NO auto shared LANs -> all ready");


        if (!experiment.getKeypairs().isEmpty()) {
            LOG.info("Experiment requires a distributed keypair. Will generate and distribute it.");
            setAndRunState(jobStateFactory.createDistributeSshKeysState(this));
        }

        //TODO: detect if these 2 steps are needed
//        boolean hasWaitForSetupReady = node.getExecuteServices().stream().anyMatch(es -> es.getFinishedFile() != null);
//
//        boolean hasAnsible = ansibleService != null &&
//                (
//                        ansibleService.getInventory() != null ||
//                        ansibleService.getExecutePlaybook() != null
//                );

        //TODO: move "wait for setup script ready" out of ExecuteAnsibleServicesState
        setAndRunState(jobStateFactory.createExecuteAnsibleServicesState(this));



        if (experiment.getExperimentSpecification() != null) {
            LOG.info("Experiment run with ExperimentSpecification. Will now start its dir, upload and execute phases.");

            experiment.getExperimentSpecificationLogger().firePostRSpec(
                    experiment.getExperimentSpecification().getRspecs().get(0), true);
            experiment.getExperimentSpecificationLogger().firePostRspecAll(
                    true, getExperimentClientIds());

            List<State> states = jobStateFactory.createOrderedExperimentServicesStates(this);
            for (State state: states) {
                setAndRunState(state);
                if (state.getStatus() != ExperimentTaskStatus.SUCCESS) {
                    //INACTIVE, BUSY, FAILED, WARNING
                    experiment.getExperimentSpecificationLogger().firePostExecuteAll(false);
                    success = false;
                    return false;
                }
            }
            experiment.getExperimentSpecificationLogger().firePostExecuteAll(true);
        }

        LOG.debug("SetupSoftwareExperimentJob stopped successfully");

        success = true;
        return true;
    }

    @Nonnull
    private List<String> getExperimentClientIds() {
        if (experiment.getSliceOrNull() == null || experiment.getSliceOrNull().getRequestRspec() == null) {
            return Collections.emptyList();
        }
        RequestRspecSource requestRspecSource = experiment.getSliceOrNull().getRequestRspec();
        FXModelRspec model = requestRspecSource == null ?
                null :
                requestRspecSource.getModelRspec(ModelRspecType.FX);
        assert model != null;

        if (model == null) {
            return Collections.emptyList();
        }

        return model.getNodes().stream().map(RspecNode::getClientId).filter(Objects::nonNull).collect(Collectors.toList());
    }

    public Boolean isSuccess() {
        return success;
    }

    public class AutoShareLanPartsState extends SlicedState<AutoShareLanPartStateSlice> {
        private final List<AutoShareLanPartStateSlice> slices;

        protected AutoShareLanPartsState() {
            super("Auto Sharing LANs in all ExperimentParts");

            slices = new ArrayList<>();

            List<SfaExperimentPart> sfaExperimentParts =
                    experiment.getPartsStream()
                            .filter(part -> part.getState() != InternalState.FAILED)
                            .filter(part -> part instanceof SfaExperimentPart)
                            .map(part -> (SfaExperimentPart) part)
                            .collect(Collectors.toList());

            if (experiment.getSliceOrNull() == null)
                throw new IllegalStateException("experiment.getSlice() is null which is not supposed to occur in this state");
            if (experiment.getSliceOrNull().getRequestRspec() == null)
                throw new IllegalStateException("experiment.getSlice().getRequestRspec() is null which is not supposed to occur in this state");
            if (experiment.getSliceOrNull().getRequestRspec().getModelRspec() == null)
                throw new IllegalStateException("experiment.getSlice().getRequestRspec().getModelRspec() is null which is not supposed to occur in this state");
            if (experiment.getSliceOrNull().getRequestRspec().getModelRspec().getLinks() == null)
                throw new IllegalStateException("experiment.getSlice().getRequestRspec().getModelRspec().getLinks() is null which is not supposed to occur in this state");

            Collection<? extends RspecLink> links = experiment.getSliceOrNull().getRequestRspec().getModelRspec().getLinks();

            for (SfaExperimentPart sfaExperimentPart : sfaExperimentParts) {
                Server auth = sfaExperimentPart.getConnectSfaAuthority();
                boolean hasAutoShareLans = false;

                for (RspecLink link : links) {
                    List<String> autoShareLanNames = link.getAutoShareLanNames();
                    //noinspection ConstantConditions
                    if (autoShareLanNames != null && !autoShareLanNames.isEmpty() && link.getComponentManagerUrns().contains(auth.getDefaultComponentManagerAsGeniUrn())) {
                        hasAutoShareLans = true;
                        break;
                    }
                }

                if (hasAutoShareLans) {
                    LOG.debug("Auto Sharing LAN(s) in " + auth.getName());
                    AutoShareLanPartStateSlice slice = new AutoShareLanPartStateSlice(sfaExperimentPart);
                    slices.add(slice);
                } else {
                    LOG.debug("No Auto Shared LAN  in " + auth.getName());
                }
            }

            LOG.debug("AutoShareLanPartsState constructor finished with " + slices.size() + " slices.");

            assert !slices.isEmpty();
        }

        @Override
        public Collection<AutoShareLanPartStateSlice> getSlices() {
            return slices;
        }
    }

    public class AutoShareLanPartStateSlice extends ExperimentPartStateSlice<SfaExperimentPart> {

        public AutoShareLanPartStateSlice(SfaExperimentPart experimentPart) {
            super(SetupSoftwareExperimentJob.this, experimentPart);
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            if (!supportsShareLan(experimentPart.getConnectSfaAuthority())) {
                LOG.warn("Share LAN on {} is not supported. Will try anyway.", experimentPart.getName());
                //return ExperimentTaskStatus.WARNING;
            }
            if (experiment.getSliceOrNull() == null)
                throw new IllegalStateException("experiment.getSlice() is null which is not supposed to occur in this state");
            if (experiment.getSliceOrNull().getRequestRspec() == null)
                throw new IllegalStateException("experiment.getSlice().getRequestRspec() is null which is not supposed to occur in this state");
            if (experiment.getSliceOrNull().getRequestRspec().getModelRspec() == null)
                throw new IllegalStateException("experiment.getSlice().getRequestRspec().getModelRspec() is null which is not supposed to occur in this state");
            if (experiment.getSliceOrNull().getRequestRspec().getModelRspec().getLinks() == null)
                throw new IllegalStateException("experiment.getSlice().getRequestRspec().getModelRspec().getLinks() is null which is not supposed to occur in this state");

            Collection<? extends RspecLink> links = experiment.getSliceOrNull().getRequestRspec().getModelRspec().getLinks();

            ExperimentTaskStatus worstStatus = null;
            Server auth = experimentPart.getConnectSfaAuthority();
            LOG.debug("Start sharing LAN(s) on " + auth.getDefaultComponentManagerAsGeniUrn());
            for (RspecLink link : links) {
                List<String> autoShareLanNames = link.getAutoShareLanNames();
                //noinspection ConstantConditions
                if (autoShareLanNames != null && !autoShareLanNames.isEmpty() && link.getComponentManagerUrns().contains(auth.getDefaultComponentManagerAsGeniUrn())) {
                    String sharedLanName = autoShareLanNames.get(0);

                    if (autoShareLanNames.size() > 1)
                        LOG.warn("There is more than one auto share LAN name per link (for link\"" + link.getClientId() + "\"), ignoring all but first.");

                    LOG.debug("Sharing LAN " + link.getClientId() + " as " + sharedLanName + " on " + auth.getDefaultComponentManagerAsGeniUrn());

                    ShareLanAtAuthorityPartState shareLanState =
                            jobStateFactory.createShareLanState(SetupSoftwareExperimentJob.this, experimentPart, link.getClientId(), sharedLanName);

                    setAndRunState(shareLanState);
                    ExperimentTaskStatus status = shareLanState.getStatus();
                    LOG.debug("Shared LAN " + link.getClientId() + " as " + sharedLanName + " on " + auth.getDefaultComponentManagerUrn() + " -> status = " + status);

                    if (status != ExperimentTaskStatus.SUCCESS || worstStatus == null) {
                        worstStatus = status;
                    }
                }
            }

            //there should be at least one link shared, or this shouldn't have been called.
            assert worstStatus != null;

            LOG.debug("Stopped sharing LAN(s) on " + auth.getDefaultComponentManagerUrn() + " -> worst status = " + worstStatus);

            return worstStatus;
        }

        private boolean supportsShareLan(Server auth) {
            AggregateManagerWrapper amWrapper =
                    aggregateManagerWrapperFactory.getAggregateManagerWrapper(auth);

            return amWrapper.isShareLanSupported();
        }
    }






    private Collection<SfaExperimentPart> getSfaParts() {
        return experiment.getPartsStream()
                .filter(part -> part instanceof SfaExperimentPart)
                .map(SfaExperimentPart.class::cast)
                .collect(Collectors.toList());

    }

    private Collection<NonSfaExperimentPartController> getNonSfaExperimentPartControllers() {

        Set<NonSfaExperimentPartController> controllers = new HashSet<>();

        for (ExperimentPart experimentPart : experiment.getPartsListCopy()) {
            if (experimentPart instanceof SfaExperimentPart)
                continue;


            ExperimentPartController epc = epcManager.getController(experimentPart);
            if (!(epc instanceof NonSfaExperimentPartController))
                throw new RuntimeException("Expected a non-sfa ExperimentPartController");

            controllers.add((NonSfaExperimentPartController) epc);
        }

        return controllers;
    }

    private boolean hasNonSfaParts() {
        return experiment.getPartsStream().anyMatch(part -> !(part instanceof SfaExperimentPart));
    }

    private boolean hasSfaParts() {
        return experiment.getPartsStream().anyMatch(part -> (part instanceof SfaExperimentPart));
    }
}
